"""Model commands"""

import click

from constellaxion.handlers.cloud_job import AWSDeployJob, GCPDeployJob
from constellaxion.ui.server.run import PromptManager
from constellaxion.utils import get_job


@click.group()
def model():
    """Manage model jobs"""
    pass


@model.command()
def prompt():
    """Prompt a deployed model"""
    click.clear()  # Clear the screen
    prompt_manager = PromptManager()
    prompt_manager.run()


@model.command()
def train():
    """Run training job"""
    click.echo(click.style("Preparing training job...", fg="blue"))
    config = get_job()
    if config:
        cloud = config.get("deploy", {}).get("provider", None)
        if cloud == "gcp":
            job = GCPDeployJob()
            job.run(config)
        # elif cloud == "aws":
        #     job = AWSDeployJob()
        #     job.run(config)


@model.command(help="Serve a trained model")
def serve():
    """Serve Model"""
    config = get_job()
    if config:
        model_id = config.get("model", {}).get("model_id", None)
        click.echo(click.style(f"Serving model with ID: {model_id}", fg="blue"))
        cloud = config.get("deploy", {}).get("provider", None)
        if cloud == "gcp":
            job = GCPDeployJob()
            job.serve(config)
        elif cloud == "aws":
            job = AWSDeployJob()
            job.serve(config)


@model.command()
def deploy():
    """Deploy a model"""
    click.echo(click.style("Deploying model...", fg="blue"))
    job_config = get_job()
    cloud = job_config.get("deploy", {}).get("provider", None)
    if cloud == "gcp":
        job = GCPDeployJob()
        job.deploy(job_config)
    elif cloud == "aws":
        job = AWSDeployJob()
        job.deploy(job_config)


@model.command()
def view():
    """View the status or details of one or more jobs"""
    get_job(show=True)



"""AWS LMI deployment module for deploying models to SageMaker endpoints using Large Model Inference."""

import boto3
import sagemaker
from sagemaker.djl_inference.model import DJLModel

from constellaxion.services.aws.utils import get_aws_account_id
from constellaxion.utils import get_model_map


def create_model_from_lmi_container(
    base_model: str, env_vars: dict, execution_role: str
):
    """Creates a SageMaker model using the LMI container."""
    model = DJLModel(model_id=base_model, env=env_vars, role=execution_role)
    return model


def deploy_model_to_endpoint(model, model_id: str, instance_type: str):
    """Deploys a model to a SageMaker endpoint using LMI."""
    endpoint_name = sagemaker.utils.name_from_base(model_id)
    predictor = model.deploy(
        initial_instance_count=1,
        instance_type=instance_type,
        endpoint_name=endpoint_name,
    )
    return predictor


def run_aws_deploy_job(config):
    """Runs the LMI deployment job by creating and deploying a model to SageMaker."""
    deploy_config = config.get("deploy", {})
    if not deploy_config:
        raise KeyError("Invalid config, missing deploy section")
    model_config = config.get("model", {})
    if not model_config:
        raise KeyError("Invalid config, missing model section")
    base_model_alias = model_config.get("base_model", None)
    model_id = model_config.get("model_id", None)
    hf_token = model_config.get("hf_token", None)
    region = deploy_config.get("region", None)
    iam_role = deploy_config.get("iam_role", None)
    account_id = get_aws_account_id()
    role_arn = f"arn:aws:iam::{account_id}:role/{iam_role}"
    # Get the model infra config from the constellaxion database
    model_map = get_model_map(base_model_alias)
    base_model = model_map.get("base_model", None)
    hf_token_required = model_map.get("hf_token_required", False)
    if hf_token_required and not hf_token:
        raise ValueError(
            "This is a protected model, please provide a valid HF token in model.yaml file"
        )
    max_model_length = model_map.get("max_model_length", None)
    # Use the LMI container image
    # image_uri = (
    #     "763104351884.dkr.ecr.us-west-2.amazonaws.com/"
    #     "djl-inference:0.25.0-lmi-deepspeed0.10.0-cu118"
    infra_config = model_map.get("aws_infra", {})
    instance_type = infra_config.get("instance_type", None)
    accelerator_count = infra_config.get("accelerator_count", 1)
    dtype = "float16" if not infra_config.get("dtype") else infra_config.get("dtype")
    option_rolling_batch = infra_config.get("option_rolling_batch", "vllm")

    # LMI specific environment variables
    env_vars = {
        "MODEL_ID": base_model,
        "DTYPE": dtype,
        "OPTION_MODEL_LOADING_TIMEOUT": "3600",
        "OPTION_ROLLING_BATCH": option_rolling_batch,
        "OPTION_MAX_ROLLING_BATCH_SIZE": "32",
        "OPTION_TENSOR_PARALLEL_DEGREE": str(accelerator_count),
        "OPTION_LOAD_IN_8BIT": "true" if dtype == "int8" else "false",
        "OPTION_LOAD_IN_4BIT": "true" if dtype == "int4" else "false",
    }
    if hf_token:
        env_vars["HF_TOKEN"] = hf_token
    if max_model_length:
        env_vars["OPTION_MAX_MODEL_LEN"] = str(max_model_length)
        env_vars["OPTION_MAX_SEQ_LEN"] = str(max_model_length)
    boto3.setup_default_session(region_name=region)

    # Register the model with LMI container
    model = create_model_from_lmi_container(base_model, env_vars, role_arn)

    # Deploy to endpoint
    predictor = deploy_model_to_endpoint(model, model_id, instance_type)
    endpoint_name = predictor.endpoint_name
    return endpoint_name



from google.cloud import aiplatform

from constellaxion.utils import get_model_map


def create_model_from_custom_container(model_name: str, image_uri: str, env_vars: dict):
    """Creates a Vertex AI model from a custom container image."""
    print("Creating model from custom container...")

    # Define the container and model
    model = aiplatform.Model.upload(
        display_name=model_name,
        serving_container_image_uri=image_uri,
        serving_container_ports=[8080],  # Port your container exposes
        # Optional, adjust if your container provides health checks
        serving_container_health_route="/health",
        # Optional, adjust if your container provides a prediction endpoint
        serving_container_predict_route="/predict",
        serving_container_environment_variables=env_vars,
    )
    print(f"Model created successfully: {model.resource_name}")
    return model


# Deploy the model to an endpoint


def deploy_model_to_endpoint(
    model,
    model_id: str,
    machine_type: str,
    accelerator_type: str,
    accelerator_count: int,
    replica_count: int,
    service_account: str,
):
    """Deploys a model to a Vertex AI endpoint with specified compute resources."""
    # Check if the endpoint exists, create it if not
    endpoints = aiplatform.Endpoint.list(filter=f'display_name="{model_id}"')
    if endpoints:
        endpoint = endpoints[0]  # Use the first matching endpoint
        print(f"Using existing endpoint: {endpoint.display_name}")
    else:
        # Create a new endpoint
        endpoint = aiplatform.Endpoint.create(display_name=model_id)
        print(f"Created new endpoint: {endpoint.display_name}")

    # Deploy the model to the endpoint
    model.deploy(
        endpoint=endpoint,
        deployed_model_display_name=model_id,
        traffic_split={"0": 100},  # Route all traffic to this model
        accelerator_type=accelerator_type,
        accelerator_count=accelerator_count,
        machine_type=machine_type,
        min_replica_count=replica_count,
        max_replica_count=replica_count,
        service_account=service_account,
    )

    print(f"Model deployed to endpoint: {endpoint.resource_name}")
    return endpoint.resource_name


def run_gcp_deploy_job(config):
    """Runs the GCP deployment job by creating and deploying a model to Vertex AI."""
    deploy_config = config.get("deploy", {})
    if not deploy_config:
        raise KeyError("Invalid config, missing deploy section")
    model_config = config.get("model", {})
    if not model_config:
        raise KeyError("Invalid config, missing model section")
    project_id = deploy_config.get("project_id", None)
    region = deploy_config.get("region", None)
    base_model_alias = model_config.get("base_model", None)
    model_id = model_config.get("model_id", None)
    hf_token = model_config.get("hf_token", None)
    service_account = deploy_config.get("service_account", None)
    # Get the model infra config from the constellaxion database
    model_map = get_model_map(base_model_alias)
    base_model = model_map.get("base_model", None)
    infra_config = model_map.get("gcp_infra", {})
    hf_token_required = model_map.get("hf_token_required", False)
    if hf_token_required and not hf_token:
        raise ValueError(
            "This is a protected model, please provide a valid HF token in model.yaml file and rerun `constellaxion init`"
        )
    image_uri = infra_config.get("images", {}).get("serve", None)
    machine_type = infra_config.get("machine_type", None)
    accelerator_type = infra_config.get("accelerator_type", None)
    accelerator_count = infra_config.get("accelerator_count", None)
    replica_count = infra_config.get("replica_count", None)
    dtype = infra_config.get("dtype", None)
    env_vars = {"MODEL_NAME": base_model, "DTYPE": dtype}
    if hf_token:
        env_vars["HF_TOKEN"] = hf_token
    # Initialize the Vertex AI SDK
    aiplatform.init(project=project_id, location=region)
    # # Download the model
    model = create_model_from_custom_container(base_model, image_uri, env_vars)
    # Deploy model to endpoint
    endpoint_path = deploy_model_to_endpoint(
        model,
        model_id,
        machine_type,
        accelerator_type,
        accelerator_count,
        replica_count,
        service_account,
    )
    return endpoint_path



from google.cloud import run_v2
from google.cloud.run_v2 import Service
from google.cloud.run_v2.types import Container, EnvVar, ResourceRequirements

from constellaxion.utils import get_model_map


def create_cloud_run_service(
    project_id: str,
    region: str,
    service_name: str,
    image_uri: str,
    env_vars: dict,
    service_account: str,
    memory: str = "4Gi",
    cpu: str = "2",
    gpu_type: str = "nvidia-tesla-t4",
    gpu_count: int = 1,
    min_instances: int = 1,
    max_instances: int = 10,
):
    """Creates or updates a Cloud Run service for model serving with GPU support."""
    print("Creating/updating Cloud Run service with GPU support...")

    # Initialize the Cloud Run client
    client = run_v2.ServicesClient()

    # Convert env_vars dict to list of EnvVar objects
    env_vars_list = [EnvVar(name=key, value=value) for key, value in env_vars.items()]

    # Create the container configuration
    container = Container(
        image=image_uri,
        env=env_vars_list,
        ports=[{"container_port": 8080}],  # Default port for model serving
        resources=ResourceRequirements(
            limits={
                "memory": memory,
                "cpu": cpu,
                "nvidia.com/gpu": str(gpu_count),  # Specify GPU count
            }
        ),
    )

    # Create the service configuration
    service = Service(
        name=f"projects/{project_id}/locations/{region}/services/{service_name}",
        template=run_v2.RevisionTemplate(
            containers=[container],
            service_account=service_account,
            scaling=run_v2.RevisionScaling(
                min_instance_count=min_instances,
                max_instance_count=max_instances,
            ),
            execution_environment="EXECUTION_ENVIRONMENT_GEN2",
            # Add GPU configuration
            annotations={
                "run.googleapis.com/gpu-type": gpu_type,
                "run.googleapis.com/gpu-count": str(gpu_count),
            },
        ),
    )

    # Create or update the service
    operation = client.update_service(service=service)
    result = operation.result()

    print(f"Cloud Run service created/updated successfully: {result.name}")
    return result.name


def run_gcp_cloud_run_deploy(config):
    """Runs the GCP Cloud Run deployment job for model serving with GPU support."""
    deploy_config = config.get("deploy", {})
    if not deploy_config:
        raise KeyError("Invalid config, missing deploy section")
    model_config = config.get("model", {})
    if not model_config:
        raise KeyError("Invalid config, missing model section")

    # Extract configuration values
    project_id = deploy_config.get("project_id", None)
    region = deploy_config.get("region", None)
    base_model_alias = model_config.get("base_model", None)
    model_id = model_config.get("model_id", None)
    hf_token = model_config.get("hf_token", None)
    service_account = deploy_config.get("service_account", None)

    # Get the model infra config from the constellaxion database
    model_map = get_model_map(base_model_alias)
    base_model = model_map.get("base_model", None)
    infra_config = model_map.get("gcp_infra", {})
    hf_token_required = model_map.get("hf_token_required", False)

    if hf_token_required and not hf_token:
        raise ValueError(
            "This is a protected model, please provide a valid HF token in model.yaml file and rerun `constellaxion init`"
        )

    # Get infrastructure configuration
    image_uri = infra_config.get("images", {}).get("serve", None)
    memory = infra_config.get(
        "memory", "16Gi"
    )  # Increased default memory for GPU workloads
    cpu = infra_config.get("cpu", "4")  # Increased default CPU for GPU workloads
    gpu_type = infra_config.get("gpu_type", "nvidia-tesla-t4")
    gpu_count = infra_config.get("gpu_count", 1)
    min_instances = infra_config.get("min_instances", 1)
    max_instances = infra_config.get("max_instances", 10)
    dtype = infra_config.get("dtype", None)

    # Prepare environment variables
    env_vars = {
        "MODEL_NAME": base_model,
        "DTYPE": dtype,
        "CUDA_VISIBLE_DEVICES": "0",  # Enable GPU visibility
    }
    if hf_token:
        env_vars["HF_TOKEN"] = hf_token

    # Deploy to Cloud Run
    service_path = create_cloud_run_service(
        project_id=project_id,
        region=region,
        service_name=model_id,
        image_uri=image_uri,
        env_vars=env_vars,
        service_account=service_account,
        memory=memory,
        cpu=cpu,
        gpu_type=gpu_type,
        gpu_count=gpu_count,
        min_instances=min_instances,
        max_instances=max_instances,
    )

    return service_path
